from __future__ import annotations

from pathlib import Path

from .helpers import run


PROJECT_ROOT = Path(__file__).parent.parent.parent


def test_history():
    readme_path = PROJECT_ROOT / 'README.md'
    cmd = ['history', '--baseline-path', str(readme_path), '--no-color']
    actual = run(cmd)
    exp = '2022-09-01 11:45:28+02:00  60 =   3   -1  +58   8fe7afd10c  git@orsinium.dev'
    assert exp in actual



from __future__ import annotations

from argparse import ArgumentParser

from ._base import Command


class Version(Command):
    """Print the version of mypy-baseline.
    """

    @staticmethod
    def init_parser(parser: ArgumentParser) -> None:
        pass

    def run(self) -> int:
        from mypy_baseline import __version__
        self.print(__version__)
        return 0



from __future__ import annotations

from collections import defaultdict

from .._error import Error
from ._base import Command


class Filter(Command):
    """Filter out old mypy errors from stdin.
    """

    def run(self) -> int:
        try:
            baseline_text = self.config.baseline_path.read_text(encoding='utf8')
        except FileNotFoundError:
            baseline_text = ''
        baseline = baseline_text.splitlines()

        unresolved_errors: list[Error] = []
        new_errors: list[Error] = []

        for line in self.stdin:
            error = Error.new(line)
            if error is None:
                self.print(line, end='')
                continue
            if self.config.is_ignored(error.message):
                continue
            clean_line = error.get_clean_line(self.config)
            try:
                baseline.remove(clean_line)
            except ValueError:
                self.print(line, end='')
                new_errors.append(error)
            else:
                unresolved_errors.append(error)

        fixed_errors: list[Error] = []
        for line in baseline:
            error = Error.new(line)
            if error is None:
                print(f'invalid baseline, cannot parse line: {line}')
                return 1
            fixed_errors.append(error)

        fixed_count = len(fixed_errors)
        new_count = len(new_errors)
        unresolved_count = len(unresolved_errors)

        # calculate exit code
        if not fixed_count and not new_count and not unresolved_count:
            return 0
        exit_code = new_count
        if not self.config.allow_unsynced:
            exit_code += fixed_count
        if exit_code > 100:
            exit_code = 100
        if self.config.hide_stats:
            return exit_code

        # print short summary
        self.print()
        self.print('total errors:')
        self.print(f'  fixed: {self.colors.green(fixed_count)}')
        self.print(f'  new: {self.colors.red(new_count)}')
        self.print(f'  unresolved: {self.colors.blue(unresolved_count)}')
        self.print()

        # print stats for each error code (category)
        stats_total: defaultdict[str, int] = defaultdict(int)
        stats_fixed: defaultdict[str, int] = defaultdict(int)
        stats_new: defaultdict[str, int] = defaultdict(int)
        for error in fixed_errors:
            stats_total[error.category] += 1
            stats_fixed[error.category] += 1
        for error in new_errors:
            stats_total[error.category] += 1
            stats_new[error.category] += 1
        for error in unresolved_errors:
            stats_total[error.category] += 1
        self.print('errors by error code:')
        sorted_stats = sorted(
            stats_total.items(),
            key=lambda x: x[::-1],
            reverse=True,
        )
        for category, total in sorted_stats:
            total_formatted = f'{total: >3}'
            line = f'  {category:24} {self.colors.blue(total_formatted)}'
            fixed = stats_fixed[category]
            if fixed:
                fixed_formatted = f'{-fixed: >3}'
                line += f' {self.colors.green(fixed_formatted)}'
            new = stats_new[category]
            if new:
                new_formatted = f'{new: >+3}'
                line += f' {self.colors.red(new_formatted)}'
            self.print(line)

        msg = self.colors.get_exit_message(fixed=fixed_count, new=new_count)
        self.print(msg)
        return exit_code



from __future__ import annotations

from io import StringIO

from mypy_baseline import main


LINE1 = 'views.py:69: error: Hello world  [assignment]\r\n'
LINE2 = 'settings.py:42: error: How are you?  [union-attr]\r\n'


def run(cmd: list[str], exit_code: int = 0) -> str:
    stdout = StringIO()
    code = main(cmd, StringIO(), stdout)
    assert code == exit_code
    stdout.seek(0)
    return stdout.read()



from __future__ import annotations

import re
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from ._config import Config

COLOR_PATTERN = '(\x1b\\[\\d*m?|\x0f)*'
REX_COLOR = re.compile(COLOR_PATTERN)
REX_LINE = re.compile(rf"""
    (?P<path>.+\.py):
    (?P<lineno>[0-9]+):\s
    {COLOR_PATTERN}(?P<severity>[a-z]+):{COLOR_PATTERN}\s
    (?P<message>.+?)
    (?:\s\s{COLOR_PATTERN}\[(?P<category>[a-z-]+)\]{COLOR_PATTERN})?
    \s*
""", re.VERBOSE | re.MULTILINE)
REX_LINE_IN_MSG = re.compile(r'defined on line \d+')


@dataclass
class Error:
    raw_line: str
    _match: re.Match[str]

    @classmethod
    def new(self, line: str) -> Error | None:
        match = REX_LINE.fullmatch(line)
        if match is None:
            return None
        return Error(line, match)

    @cached_property
    def path(self) -> Path:
        return Path(self._match.group('path'))

    @cached_property
    def line_number(self) -> int:
        result = int(self._match.group('lineno'))
        assert result >= 0
        return result

    @cached_property
    def severity(self) -> str:
        return self._match.group('severity')

    @cached_property
    def message(self) -> str:
        return self._match.group('message')

    @cached_property
    def category(self) -> str:
        return self._match.group('category') or 'note'

    def get_clean_line(self, config: Config) -> str:
        path = Path(*self.path.parts[:config.depth])
        pos = self.line_number if config.preserve_position else 0
        msg = REX_COLOR.sub('', self.message).strip()
        msg = REX_LINE_IN_MSG.sub('defined on line 0', msg)
        line = f'{path}:{pos}: {self.severity}: {msg}'
        if self.category != 'note':
            line += f'  [{self.category}]'
        return line



from __future__ import annotations

from io import StringIO

from mypy_baseline import main

from .helpers import LINE1, LINE2, run


def test_filter():
    stdin = StringIO()
    stdin.write(LINE1)
    stdin.write(LINE2)
    stdin.seek(0)
    stdout = StringIO()
    code = main(['filter'], stdin, stdout)
    assert code == 2
    stdout.seek(0)
    actual = stdout.read()
    assert LINE1.strip() in actual
    assert LINE2.strip() in actual
    assert '  assignment  ' in actual
    assert '  union-attr  ' in actual
    assert '  unresolved' in actual
    assert 'Your changes introduced' in actual


def test_filter__empty_stdin():
    actual = run(['filter'])
    assert actual == ''



from __future__ import annotations

from io import StringIO
from pathlib import Path

from mypy_baseline import main

from .helpers import LINE1, LINE2


def test_sync(tmp_path: Path):
    blpath = tmp_path / 'bline.txt'
    stdin = StringIO()
    stdin.write(LINE1)
    stdin.write(LINE2)
    stdin.seek(0)
    code = main(['sync', '--baseline-path', str(blpath)], stdin, StringIO())
    assert code == 0
    actual = blpath.read_text()
    line1, line2 = actual.splitlines()
    assert line1 == 'views.py:0: error: Hello world  [assignment]'
    assert line2 == 'settings.py:0: error: How are you?  [union-attr]'
